unit Builtin;

interface

uses Windows, NLib, NUtil, NSock, Sntpc, Smart, SmartHelper, Gauge;

const
  MaxCpu = 32;

type
  TPlugin = class(TGauge)
    Handle: THandle;
    PluginName: string;
    InitProc: function: Integer; stdcall;
    UpdateProc: procedure; stdcall;
    UnloadProc: procedure; stdcall;
    constructor CreateByPlugin(const ASymbols: TStrDict;
      const ALogger: TLogger; const APluginName: string);
    destructor Destroy; override;
    procedure Init; override;
    procedure Update; override;
  end;

  TTime = class(TGauge)
    Server: string;
    PrevSendTime: Integer;
    SyncInterval: Integer;
    Sntpc: TSntpc;
    DayNamesL: array [0..6] of string;
    destructor Destroy; override;
    procedure Init; override;
    procedure Update; override;
    procedure OnSynced(const SyncInfo: TSntpcSyncInfo);
    procedure OnError(Code: Integer);
  end;

  TMem = class(TGauge)
    procedure Init; override;
    procedure Update; override;
  end;

  TUptime = class(TGauge)
    procedure Update; override;
  end;

  TDisk = class(TGauge)
    procedure Init; override;
    procedure Update; override;
  end;

  TCpu = class(TGauge)
    hNtdll: THandle;
    Num: Integer;
    OldWholeTime: array[0..MaxCpu-1] of Comp;
    OldUserTime: array[0..MaxCpu-1] of Comp;
    OldKernelTime: array[0..MaxCpu-1] of Comp;
    OldIdleTime: array[0..MaxCpu-1] of Comp;
    IsWin9x: Boolean;
    destructor Destroy; override;
    procedure Init; override;
    procedure Update; override;
  end;

  TNet = class(TGauge)
    hIphlpapi: THandle;
    OldInOct: DWORD;
    OldOutOct: DWORD;
    OldTime: DWORD;
    destructor Destroy; override;
    procedure Init; override;
    procedure Update; override;
  end;

  TUdp = class(TGauge)
    Udp: TNUdp;
    destructor Destroy; override;
    procedure Init; override;
    procedure OnRecv(Sender: TObject);
  end;

  TMoon = class(TGauge)
    procedure Init; override;
    procedure Update; override;
  end;

  THdtemp = class(TGauge)
    FirstUpdate: Boolean;
    EnableDrive: array[0..MAX_IDE_DRIVES-1] of Boolean;
    procedure Init; override;
    procedure Update; override;
  end;

  TProc = class(TGauge)
    hPsapi: THandle;
    destructor Destroy; override;
    procedure Init; override;
    procedure Update; override;
  end;

  TIme = class(TGauge)
    procedure Update; override;
  end;

const
  GaugeClasses: array[0..11] of TGaugeClass = (
    TPlugin, TTime, TMem, TUptime, TDisk, TCpu, TNet, TUdp, TMoon, THdtemp,
    TProc, TIme
  );

procedure SetDefaultValue(const ASymbols: TStrDict; const Name: string); 

implementation

uses Imm, NSys, NDate, MoonPhase;

procedure SetDefaultValue(const ASymbols: TStrDict; const Name: string); 
begin
  if Name = 'hostname' then
    ASymbols[Name] := GetHostName
  else
    ASymbols[Name] := '';
end;

{ plugin }

destructor TPlugin.Destroy;
begin
  if Assigned(UnloadProc) then
    UnloadProc;
  if Handle <> 0 then
    FreeLibrary(Handle);
  inherited;
end;

constructor TPlugin.CreateByPlugin(const ASymbols: TStrDict;
  const ALogger: TLogger; const APluginName: string);
begin
  PluginName := APluginName;
  Create(ASymbols, ALogger);
end;

procedure TPlugin.Init;
begin
  Handle := LoadLibrary(PChar(PluginName+'.nmp'));
  if Handle = 0 then
    Log(Self, PluginName+': can''t load')
  else begin
    Log(Self, PluginName+': loadded');
    InitProc := GetProcAddress(Handle, 'gauge_init');
    UpdateProc := GetProcAddress(Handle, 'gauge_update');
    UnloadProc := GetProcAddress(Handle, 'gauge_unload');
    if Assigned(InitProc) then
      DefaultInterval := InitProc;
  end;
end;

procedure TPlugin.Update;
begin
  if Assigned(UpdateProc) then
    UpdateProc;
end;

{ time }

destructor TTime.Destroy;
begin
  Sntpc.Free;
end;

procedure TTime.Init;
var
  i: Integer;
begin
  Server := Conf('server', 'ntp.jst.mfeed.ad.jp');
  SyncInterval := 30;
  if Conf('sync', 'on') = 'on' then
  begin
    Sntpc := TSntpc.Create;
    Sntpc.OnSynced := OnSynced;
    Sntpc.OnError := OnError;
    Log(Self, Fmt('SNTP client enabled. server=%s sync interval %s min.',
      [Server, SyncInterval]));
  end;
  for i := 0 to 6 do
    DayNamesL[(i + 8) mod 7] :=
      LocaleStr(LOCALE_USER_DEFAULT, LOCALE_SABBREVDAYNAME1+i);
end;

procedure TTime.Update;
const
  DayNames: array [0..6] of string = (
    'Sun', 'Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat'
  );
  MonthNames: array [0..11] of string = (
    'Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun',
    'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec'
  );
var
  st: TSystemTime;
  t, yd: Integer;
begin
  if Assigned(Sntpc) then
  begin
    t := GetTickCount;
    if Sntpc.Syncing and ((t-PrevSendTime) > 5000) then // time out 5 sec.
    begin
      Sntpc.Abort;
      Log(Self, 'SNTP time out.');
    end;
    if (PrevSendTime = 0) or ((t-PrevSendTime) > (SyncInterval*60*1000)) then
    begin
      PrevSendTime := t;
      Sntpc.Sync(Server);
    end;
  end;

  GetLocalTime(St);
  with st do
  begin
    Value['year'] := IStr(wYear);
    Value['shortyear'] := IZStr(wYear mod 1000 mod 100, 2);
    Value['month'] := IZStr(wMonth, 2);
    Value['monthname'] := MonthNames[wMonth-1];
    Value['day'] := IZStr(wDay, 2);
    Value['dayname'] := DayNames[wDayOfWeek];
    Value['daynamel'] := DayNamesL[wDayOfWeek];
    Value['hour'] := IZStr(wHour, 2);
    Value['min'] := IZStr(wMinute, 2);
    Value['sec'] := IZStr(wSecond, 2);
    Value['weeknum'] := IZStr(ISOWeekNumber(wYear, wMonth, wDay), 2);
    yd := Yearday(wYear, wMonth, wDay);
    Value['yearday'] := IStr(yd);
    Value['yeardayremain'] := IStr(365+Ord(IsLeap(wYear)) - yd);
  end;
end;

procedure TTime.OnSynced(const SyncInfo: TSntpcSyncInfo);
var
  adj: string;
begin
  with SyncInfo do
  begin
    if Adjusted then
      adj := ' adjust time.'
    else
      adj := '';
    Log(Self, Fmt('server %s, stratum %s, offset %s, delay %s%s',
      [IpAddr, Stratum, FStr(Offset, 6), FStr(Delay, 5), adj]));
  end;
end;

procedure TTime.OnError(Code: Integer);
begin
  case Code of
    SNTPC_ERR_SYNC: Log(Self, 'SNTP sync error.');
    SNTPC_ERR_SEND: Log(Self, 'SNTP send error.');
    SNTPC_ERR_HOSTNOTFOUND: Log(Self, 'SNTP host not found.');
  end;
end;

{ mem }

procedure TMem.Init;
begin
  DefaultInterval := 3;
end;

procedure TMem.Update;
var
  stat: TMemoryStatus;
begin
  stat.dwLength := SizeOf(stat);
  GlobalMemoryStatus(stat);
  with stat do
  begin
    Value['phys.size'] := IStr(dwTotalPhys);
    Value['phys.avail'] := IStr(dwAvailPhys);
    Value['phys.used'] := IStr(dwTotalPhys - dwAvailPhys);
    Value['pagefile.size'] := IStr(dwTotalPageFile);
    Value['pagefile.avail'] := IStr(dwAvailPageFile);
    Value['pagefile.used'] := IStr(dwTotalPageFile - dwAvailPageFile);
  end;
end;

{ uptime }

procedure TUptime.Update;
var
  t: DWORD;
  c: Comp;
begin
  t := GetTickCount;
  c := 0;
  // copy to Comp since DWORD is signed at Delphi3
  Move(t, c, SizeOf(t));
  t := Trunc(c / 1000);
  Value['days'] := IStr(t div (60*60*24));
  Value['hour'] := IZStr(t mod (60*60*24) div 3600, 2);
  Value['min'] := IZStr(t mod 3600 div 60, 2);
  Value['sec'] := IZStr(t mod 60, 2);
end;

{ disk }

type
  PComp = ^Comp;
var
  GetDiskFreeSpaceEx: function (lpDirectoryName: PChar;
    lpFreeBytesAvailableToCaller, lpTotalNumberOfBytes,
    lpTotalNumberOfFreeBytes: PComp): BOOL; stdcall;

procedure TDisk.Init;
begin
  DefaultInterval := 5;
  GetDiskFreeSpaceEx := GetProcAddress(GetModuleHandle('kernel32.dll'),
    'GetDiskFreeSpaceExA');
end;

procedure TDisk.Update;
var
  avail, size: Comp;
  SectorsPerCluster, BytesPerSector, FreeClusters, Clusters: DWORD;
  drv: Char;
  r: Boolean;
begin
  for drv := 'a' to 'z' do
    if Exists(drv + '.size') or Exists(drv + '.avail') or Exists(drv + '.used') then
    begin
      if Assigned(GetDiskFreeSpaceEx) then
        r := GetDiskFreeSpaceEx(PChar(drv+':\'), @avail, @size, nil)
      else begin
        // for Windows95 OSR1
        r := GetDiskFreeSpace(PChar(drv+':\'), SectorsPerCluster,
          BytesPerSector, FreeClusters, Clusters);
        avail := FreeClusters*SectorsPerCluster*BytesPerSector;
        size := Clusters*SectorsPerCluster*BytesPerSector;
      end;
      if r then
      begin
        Value[drv + '.size'] := CStr(size);
        Value[drv + '.avail'] := CStr(avail);
        Value[drv + '.used'] := CStr(size-avail);
      end;
    end;
end;

{ cpu }

const
  SystemProcessorTimes = 8;

type
  TSystemProcessorTimes = record
    liIdleTime: TLargeInteger;
    liKernelIdleTime: TLargeInteger;
    liUserOtherTime: TLargeInteger;
    dwSpare: array[0..5] of DWORD;
  end;

var
  NtQuerySystemInformation: function (SystemInformationClass: UINT;
    SystemInformation: Pointer; SystemInformationLength: ULONG;
    ReturnLength: PULONG): Longint; stdcall;

function AccessWin9xCPUUsage(const SubKey: string): DWORD;
var
  size: DWORD;
  hkey1, hkey2: HKEY;
begin
  Result := 0;
  size := SizeOf(Result);
  if RegOpenKeyEx(HKEY_DYN_DATA, 'PerfStats', 0, KEY_READ,
      hkey1) = ERROR_SUCCESS then
  begin
    if RegOpenKeyEx(hkey1, PChar(SubKey), 0, KEY_READ,
      hkey2) = ERROR_SUCCESS then
    begin
      RegQueryValueEx(hkey2, 'KERNEL\CPUUsage', nil, nil, @Result, @size);
      RegCloseKey(hkey2);
    end;
    RegCloseKey(hkey1);
  end;
end;

procedure TCpu.Init;
var
  si: TSystemInfo;
  oi: TOsVersionInfo;
begin
  oi.dwOSVersionInfoSize := SizeOf(oi);
  GetVersionEx(oi);
  IsWin9x := oi.dwPlatformId = VER_PLATFORM_WIN32_WINDOWS;
  if IsWin9x then
  begin
    Num := 1;
    AccessWin9xCPUUsage('StartStat');
  end else
  begin
    GetSystemInfo(si);
    Num := si.dwNumberOfProcessors;
    hNtdll := LoadLibrary('ntdll.dll');
    if hNtdll <> 0 then
      NtQuerySystemInformation :=
        GetProcAddress(hNtdll, 'NtQuerySystemInformation');
  end;
end;

destructor TCpu.Destroy;
begin
  if IsWin9x then
    AccessWin9xCPUUsage('StopStat')
  else begin
    if hNtDll <> 0 then
      FreeLibrary(hNtdll);
    NtQuerySystemInformation := nil;
  end;
end;

procedure TCpu.Update;
var
  spi: array[0..MaxCpu-1] of TSystemProcessorTimes;
  t, wholetime, usertime, kerneltime, idletime: Comp;
  load, kernel, user, idle: Integer;
  loadtotal, kerneltotal, usertotal, idletotal: Integer;
  i: Integer;
begin
  if IsWin9x then
  begin
    load := AccessWin9xCPUUsage('StatData');
    Value['p0.load'] := IStr(load);
    Value['p0.idle'] := IStr(100-load);
    Value['load'] := IStr(load);
    Value['idle'] := IStr(100-load);
  end else
  begin
    if not Assigned(NtQuerySystemInformation) then
      Exit;

    NtQuerySystemInformation(SystemProcessorTimes, @spi, SizeOf(spi), nil);

    loadtotal := 0;
    usertotal := 0;
    kerneltotal := 0;
    idletotal := 0;
    for i := 0 to Num-1 do
    begin
      t := PComp(@spi[i].liKernelIdleTime)^ + PComp(@spi[i].liUserOtherTime)^;
      wholetime := t - OldWholeTime[i];
      OldWholeTime[i] := t;

      usertime := PComp(@spi[i].liUserOtherTime)^ - OldUserTime[i];
      OldUserTime[i] := PComp(@spi[i].liUserOtherTime)^;

      t := PComp(@spi[i].liKernelIdleTime)^ - PComp(@spi[i].liIdleTime)^;
      kerneltime := t - OldKernelTime[i];
      OldKernelTime[i] := t;

      idletime := PComp(@spi[i].liidleTime)^ - OldIdleTime[i];
      OldIdleTime[i] := PComp(@spi[i].liidleTime)^;

      if wholetime = 0 then
      begin
        load := 0;
        kernel := 0;
        user := 0;
        idle := 0;
      end
      else begin
        load := Trunc(100 - (idletime * 100) / wholetime);
        kernel := Trunc(kerneltime * 100 / wholetime);
        user := Trunc(usertime * 100 / wholetime);
        idle := Trunc(idletime * 100 / wholetime);
      end;

      Value['p' + IStr(i) + '.load'] := IStr(load);
      Value['p' + IStr(i) + '.load.kernel'] := IStr(kernel);
      Value['p' + IStr(i) + '.load.user'] := IStr(user);
      Value['p' + IStr(i) + '.idle'] := IStr(idle);
      Inc(loadtotal, load);
      Inc(kerneltotal, kernel);
      Inc(usertotal, user);
      Inc(idletotal, idle);
    end;
    Value['load'] := IStr(loadtotal div Num);
    Value['load.kernel'] := IStr(kerneltotal div Num);
    Value['load.user'] := IStr(usertotal div Num);
    Value['idle'] := IStr(idletotal div Num);
  end;
end;

{ net }

const
  MAX_INTERFACE_NAME_LEN = 256;
  MAXLEN_IFDESCR = 256;
  MAXLEN_PHYSADDR = 8;

type
  PMibIfRow = ^TMibIfRow;
  TMibIfRow = record
    wszName: array [0..MAX_INTERFACE_NAME_LEN-1] of WCHAR;
    dwIndex: DWORD;
    dwType: DWORD;
    dwMtu: DWORD;
    dwSpeed: DWORD;
    dwPhysAddrLen: DWORD;
    bPhysAddr: array [0..MAXLEN_PHYSADDR-1] of Byte;
    dwAdminStatus: DWORD;
    dwOperStatus: DWORD;
    dwLastChange: DWORD;
    dwInOctets: DWORD;
    dwInUcastPkts: DWORD;
    dwInNUcastPkts: DWORD;
    dwInDiscards: DWORD;
    dwInErrors: DWORD;
    dwInUnknownProtos: DWORD;
    dwOutOctets: DWORD;
    dwOutUcastPkts: DWORD;
    dwOutNUcastPkts: DWORD;
    dwOutDiscards: DWORD;
    dwOutErrors: DWORD;
    dwOutQLen: DWORD;
    dwDescrLen: DWORD;
    bDescr: array[0..MAXLEN_IFDESCR-1] of Byte;
  end;
  PMibIfTable = ^TMibIfTable;
  TMibIfTable = record
    dwNumEntries: DWORD;
    table: array [0..0] of TMibIfRow;
  end;

var
  GetIfTable: function (pIfTable: PMibIfTable;
    var pdwSize: ULONG; bOrder: BOOL): DWORD; stdcall;

procedure TNet.Init;
const
  iphlpapi = 'iphlpapi.dll';
begin
  DefaultInterval := 2;
  hIphlpapi := LoadLibrary(iphlpapi);
  if hIphlpapi = 0 then
    Log(Self, 'cannot load '+iphlpapi)
  else
    GetIfTable := GetProcAddress(hIphlpapi, 'GetIfTable');
end;

destructor TNet.Destroy;
begin
  if hIphlpapi <> 0 then
    FreeLibrary(hIphlpapi);
  GetIfTable := nil;
end;

procedure TNet.Update;
var
  size: DWORD;
  pIfTable: PMibIfTable;
  t, InOct, OutOct: DWORD;
  i: Integer;
  td: Double;
begin
  if not Assigned(GetIfTable) then
    Exit;

  size := 0;
  InOct := 0;
  OutOct := 0;
  GetIfTable(nil, size, False);
  GetMem(pIfTable, size);
  try
    if GetIfTable(pIfTable, size, False) = NO_ERROR then
      for i := 0 to pIfTable.dwNumEntries-1 do
      begin
        Inc(InOct, pIfTable.table[i].dwInOctets);
        Inc(OutOct, pIfTable.table[i].dwOutOctets);
      end;
  finally
    FreeMem(pIfTable);
  end;
  t := GetTickCount;
  td := (t - OldTime) / 1000;
  if td = 0 then
    Exit;
  if OldTime = 0 then
  begin
    OldInOct := InOct;
    OldOutOct := OutOct;
  end;
  Value['in.bytespersec'] := IStr(Trunc((InOct - OldInOct) / td));
  Value['out.bytespersec'] := IStr(Trunc((OutOct - OldOutOct) / td));
  OldInOct := InOct;
  OldOutOct := OutOct;
  OldTime := t;
end;

{ udp }

procedure TUdp.Init;
var
  port: Integer;
begin
  port := SInt(Conf('port', '7531'));
  Udp := TNUdp.Create(port);
  Udp.OnRecv := OnRecv;
  Log(Self, 'port=' + IStr(port));
end;

destructor TUdp.Destroy;
begin
  Udp.Free;
end;

procedure TUdp.OnRecv;
var
  s: string;
  ac: Integer;
  av: array[0..0] of string;
begin
  s := Udp.Recv(1024);
  s := SplitCmd(s, ac, av);
  Value[av[0]] := s;
end;

{ moon }

procedure TMoon.Init;
begin
  DefaultInterval := 60;
end;

procedure TMoon.Update;
var
  st: TSystemTime;
  jd: Double;
begin
  GetSystemTime(st);
  jd := JulianDay(st.wYear, st.wMonth, st.wDay, st.wHour, st.wMinute);
  Value['phase'] := FStr(GetMoonPhase(jd), 1);
  Value['age'] := FStr(GetMoonAge(jd), 1);
end;

{ hdtemp }

procedure THdtemp.Init;
begin
  DefaultInterval := 30;
  FirstUpdate := True;
end;

procedure THdtemp.Update;
var
  i, j: Integer;
  d: string;
  found: Boolean;
begin
  if FirstUpdate then
  begin
    FirstUpdate := False;
    for i := 0 to MAX_IDE_DRIVES-1 do
    begin
      d := 'hd'+IStr(i);
      if not Exists(d) then
        continue;  
      try
        with TSmart.Create(i) do
          try
            if ReadIdSector then
            begin
              EnableDrive[i] := True;
              for j := High(Id.sModelNumber) downto 0 do
                if Id.sModelNumber[j] = ' ' then
                  Id.sModelNumber[j] := #0
                else
                  Break;
              Log(Self, d + ': <'+Id.sModelNumber+'> SMART enabled');
            end else
              Log(Self, d + ': can''t read SMART id sector');
          finally
            Free;
          end;
      except
        on E: EIOError do
          Log(Self, d + ': can''t open drive');
      end;
    end;
  end;

  for i := 0 to MAX_IDE_DRIVES-1 do
  begin
    if not EnableDrive[i] then
      Continue;
    d := 'hd'+IStr(i);
    try
      with TSmart.Create(i) do
        try
          if ReadAttrValues then
          begin
            found := False;
            for j := 0 to NUM_ATTRIBUTE_STRUCTS-1 do
              if Attr[j].bAttrID = ATTR_TEMPERATURE then
              begin
                Value[d] := IStr(Attr[j].bAttrValue);
                found := True;
                Break;
              end;
            if not found then
            begin
              Log(Self, d + ': no support temperture attribute');
              EnableDrive[i] := False;
            end;
          end
          else begin
            Log(Self, d + ': can''t read SMART attribute');
            EnableDrive[i] := False;
          end;
        finally
          Free;
        end;
    except
      on E: EIOError do
      begin
        Log(Self, d + ': can''t open drive');
        EnableDrive[i] := False;
      end;
    end;
  end;
end;

{ TProc }

var
  EnumProcesses: function (ProcessIds: PDWORD; Cb: DWORD;
    pBytesReturned: PDWORD): BOOL; stdcall;

procedure TProc.Init;
begin
  DefaultInterval := 2;
  hPsapi := LoadLibrary('PSAPI.DLL');
  EnumProcesses := GetProcAddress(hPsapi, 'EnumProcesses');
end;

procedure TProc.Update;
var
  id: array[0..32767] of DWORD;
  n: DWORD;
begin
  if Assigned(EnumProcesses) then
  begin
    EnumProcesses(@id, SizeOf(id), @n);
    Value['num'] := IStr(n div SizeOf(id[0]));   
  end;
end;

destructor TProc.Destroy;
begin
  if hPsapi <> 0 then
    FreeLibrary(hPsapi);
  EnumProcesses := nil;
end;

procedure TIme.Update;
const
  IMC_GETOPENSTATUS = $0005;
var
  r: Integer;
begin
  SendMessageTimeout(ImmGetDefaultIMEWnd(GetForegroundWindow),
    WM_IME_CONTROL, IMC_GETOPENSTATUS, 0, SMTO_ABORTIFHUNG, 200, r);
  if LongBool(r) then
    Value['open'] := 'on'
  else
    Value['open'] := 'off'
end;

end.
